# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Classes to assist in YAML Parsing. 
'''

import os
import yaml
import abc
import copy
from json import JSONEncoder

from arjuna.tpi.helper.arjtype import CIStringDict
from arjuna.core.error import YamlError, YamlUndefinedSectionError, YamlListIndexError

# Custom tag handlers
def join_tag(loader, node):
    seq = loader.construct_sequence(node)
    return ''.join([str(i) for i in seq])

# register tag handlers
yaml.add_constructor('!join', join_tag, Loader=yaml.SafeLoader)

class YamlElement(metaclass=abc.ABCMeta):
    '''
        Abstract Json Element. Base class for YamlList and YamlDict.
    '''

    def __init__(self, pydict_or_list):
        super().__init__()
        self.__pyobj = pydict_or_list

    def is_empty(self):
        return not self.store

    def as_str(self):
        '''
            Get string representation of this Yaml Element.
        '''
        return yaml.dump(self.__pyobj).replace("_ArDict__store:\n","")

    __str__ = as_str 
    

    def __repr__(self):
        return "Yaml as PyObj: " + str(self.__pyobj)

    def _get_item(self, name, *, strict=True, as_yaml_str=False, allow_any=False):
        val = self._get_value(name, strict=strict)
        if as_yaml_str:
            return yaml.dump(val)
        return Yaml.from_object(val, allow_any=allow_any)

    @abc.abstractmethod
    def _get_value(self, name_or_index, *, strict=True):
        pass

class YamlList(YamlElement, JSONEncoder):
    '''
        Encapsulates a list object in YAML.

        Arguments:
            pylist: Python list

        Note:
            Supports indexing just like a Python list.

            Also supports == operator. Right operand can be a Python list or a `YamlList` object.
    '''

    def __init__(self, pylist):
        self.__ylist = pylist is not None and pylist or list()
        YamlElement.__init__(self, self.__ylist)
        JSONEncoder.__init__(self)
        self.__iter = None   

    def __len__(self):
        return len(self.__ylist)

    def default(self, o):
        return self.raw_object

    @property
    def size(self):
        '''
            Number of objects in this list.
        '''
        return len(self)

    @property
    def raw_object(self):
        '''
            A copy of the Python list that this object wraps.
        '''
        return copy.deepcopy(self.__ylist)

    def __str__(self):
        return str(self.__ylist)

    def __getitem__(self, index):
        return self.__ylist[index] 

    def __iter__(self):
        self.__iter = iter(self.__ylist)
        return self

    def __next__(self):
        return Yaml.from_object(next(self.__iter), allow_any=True)

    def _get_value(self, index, *, strict=True):
        if len(self.__ylist) < index:
            return self.__ylist[index]
        else:
            if strict:
                raise YamlListIndexError(f"YamlList does not contain index: {index}. List: {self.__ylist}")
            else:
                return None 

    def __eq__(self, other):
        if type(other) is list:
            pass
        elif type(other) is YamlList:
            other = other.raw_object
        elif other is None:
            return False
        else:
            raise Exception("YamlList == operator expects list/YamlList as right operand.")

        if len(self) != len(other):
            return False
        else:
            return self.raw_object == other


class YamlDict(CIStringDict, YamlElement):
    '''
        Encapsulates Dictionary object in Json.

        Arguments:
            pydict: Python dict.

        Note:
            Supports dictionary methods as well as **.** access for key.

            Also supports == operator. Right operand can be a Python dict or a `YamlDict` object.
    '''

    def __init__(self, pydict):
        CIStringDict.__init__(self, pydict)
        YamlElement.__init__(self, self.orig_dict)
        self.__sections = tuple(self.keys())
        self.__iter = None

    @property
    def size(self):
        '''
            Number of key-value pairs in this object.
        '''
        return len(self)

    @property
    def raw_object(self):
        '''
            A copy of the Python dict that this object wraps.
        '''
        return dict(super().items())

    def __getitem__(self, key):
        return self.get_section(key, allow_any=True)

    def items(self):
        '''
            dictitems object containing key-value pairs in this object.
        '''
        return {i:self.get_section(i, allow_any=True) for i in self.orig_dict}.items()

    def __iter__(self):
        self.__iter = iter(self.orig_dict)
        return self

    def __next__(self):
        return Yaml.from_object(next(self.__iter), allow_any=True)

    def get_section(self, name, *, strict=True, as_yaml_str=False, allow_any=False):
        '''
            Get Yaml object for a section/key name.

            Arguments:
                name: Name of section/key

            Keyword Arguments:
                strict: If True, raises Exception when name is not present, else returns None. Default is True.
        '''
        val = self._get_value(name, strict=strict)
        if as_yaml_str:
            return yaml.dump(val)
        return Yaml.from_object(val, allow_any=allow_any)

    def _get_value(self, name, *, strict=True):
        if self.has_section(name):
            return super().__getitem__(name)
        else:
            if strict:
                raise YamlUndefinedSectionError(f"YamlDict object does not have a section with the name: {name}. Dict: {self.orig_dict}")
            else:
                return None

    def as_map(self):
        '''
            A copy of the Python dict that this object wraps.
        '''
        return dict(super().items())

    def has_section(self, name):
        '''
            Check if a key/section name is present. You can also use the 'in' operator instead, like a Python dict.
        '''
        return name in self

    @property
    def section_names(self):
        '''
            All section/key names in this object.
        '''
        return self.keys()

    def validate_sections_present(*section_names, atleast_one=False):
        '''
            Validate the presence of arbitrary section/key names.

            Arguments:
                section_names: arbitrary section/key names

            Keyword Arguments:
                atleast_one: If True, it is expected that atleast one of the sections is present, else all are expected to be present.
        '''
        absent_sections = []
        present_section_names = self.section_names
        for section_name in section_names:
            if section_name not in present_section_names:
                absent_sections.append(section_name)

        if len(absent_sections) == section_names or (len(absent_sections) < len(section_names) and not atleast_one):
            raise YamlUndefinedSectionError(f"Yaml object does not contains mandatory sections: {absent_sections}") 

    def __eq__(self, other):
        if type(other) is dict:
            pass
        elif type(other) is YamlDict:
            other = other.store
        elif other is None:
            return False
        else:
            raise Exception("YamlDict == operator expects dict/YamlDict as right operand. Provided {} of type {}".format(other, type(other)))

        if len(self) != len(other):
            return False
        else:
            return self.store == other

class Yaml:

    @classmethod
    def from_object(cls, yobj, *, allow_any=False):
        '''
            Convert a Python object to a YamlDict/YamlList object if applicable.

            Any iterable which is not a dict is converted to a YamlList.

            Arguments:
                yobj: dict or list or a compatible Python object.

            Keyword Arguments:
                allow_any: If True, if the object can not be coverted, same object is returned, else an Exception is raised.
        '''
        if type(yobj) is dict:
            return YamlDict(yobj)
        elif type(yobj) is list:
            return YamlList(yobj)
        elif type(yobj) is not str:
            try:
                iter(yobj)
                return YamlList(yobj)
            except:
                pass

            if allow_any:
                return yobj
            else:
                raise Exception("Unknown object type for Yaml: {}".format(type(yobj)))
        else:
            if allow_any:
                return yobj
            else:
                raise Exception("Unknown object type for Yaml: {}".format(type(yobj)))            

    @classmethod
    def from_file(cls, file_path, *, allow_any=False):
        '''
            Creates a Yaml object from file.

            Arguments:
                file_path: Absolute path of the yaml file.

            Keyword Arguments:
                allow_any: If True, if the object can not be coverted to a YamlDict/YamlList, same object is returned, else an Exception is raised.

            Returns:
                Arjuna's `YamlDict` or `YamlList` object or the same object for allow_any = True
        '''
        yaml_name = os.path.basename(file_path).split(".yaml")[0]
        f = open(file_path, "r", encoding="utf-8")
        ydict = yaml.load(f, Loader=yaml.SafeLoader)
        f.close()
        return cls.from_object(ydict, allow_any=allow_any)

    @classmethod
    def from_str(cls, ystr, *, name="yaml", allow_any=False):
        '''
            Creates a Yaml object from a YAML string.

            Arguments:
                ystr: A string representing a compatible Python object.

            Keyword Arguments:
                allow_any: If True, if the object can not be coverted to a YamlDict/YamlList, same object is returned, else an Exception is raised.

            Returns:
                Arjuna's `YamlDict` or `YamlList` object or the same object for allow_any = True
        '''
        return cls.from_object(yaml.safe_load(ystr), allow_any=allow_any)



    

    


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import abc

from arjuna.core.utils import data_utils

class UnsupportedRepresentationException(Exception):

    def __init__(self, strSourceValue, targetValueType):
        super().__init__("Value: Can not represent value >>{}<< as {}.".format(strSourceValue, targetValueType))


class Value:
    TRUES = {"YES", "TRUE", "ON", "1"}
    FALSES = {"NO", "FALSE", "OFF", "0"}

    @classmethod
    def is_set(cls, val):
        return not val.upper().strip().equals("NOT_SET")

    @classmethod
    def is_not_set(cls, val):
        return val.upper().strip().equals("NOT_SET")

    @classmethod
    def is_none(cls, val):
        return val == None

    is_null = is_none

    @classmethod
    def is_na(cls, val):
        return val.upper().strip().equals("NA")

    @classmethod
    def __wrong_repr(cls, val, target):
        raise UnsupportedRepresentationException(val, target)

    @classmethod
    def as_enum(cls, val, enum_class):
        try:
            return enum_class[val]
        except:
            cls.__wrong_repr(val, "enum constant of type " + enum_class.__name__)

    @classmethod
    def as_bool(cls, val):
        fstr = str(val)
        if fstr in cls.TRUES:
            return True
        elif fstr in cls.FALSES:
            return False
        cls.__wrong_repr(val, "boolean")

    @classmethod
    def as_number(cls, val):
        fstr = str(val)
        if re.match(r"(\-)?[0-9\.]+", fstr):
            return float(fstr)
        elif re.match("(\-)?[0-9]+", fstr):
            return int(fstr)
        else:
            cls.__wrong_repr(val, "number")

    @classmethod
    def as_int(cls, val):
        try:
            return int(float(str(val)))
        except:
            cls.__wrong_repr(val, "int")

    @classmethod
    def as_float(cls, val):
        try:
            return float(str(val))
        except:
            cls.__wrong_repr(val, "float")

    @classmethod
    def as_enum_list(cls, val, enum_class):
        try:
            if type(val) is list:
                return [enum_class[i.upper()] for i in val]
            else:
                return [cls.as_enum(val, enum_class)]
        except:
            cls.__wrong_repr(val, "enum constant list of type " + enum_class.__name__)

    @classmethod
    def as_number_list(cls, val):
        try:
            return [cls.as_number(val)]
        except:
            cls.__wrong_repr(val, "number list")
    
    @classmethod
    def as_int_list(cls, val):
        try:
            return [cls.as_int(val)]
        except:
            cls.__wrong_repr(val, "int list")
    
    @classmethod
    def as_str_list(cls, val):
        try:
            return [str(val)]
        except:
            cls.__wrong_repr(val, "string list")

    @classmethod
    def split_as_str_list(cls, val, delimiter=","):
        try:
            return data_utils.split(str(val), delimiter=delimiter)
        except:
            cls.__wrong_repr(val, "string list")




# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Provides configuration related public classes.

Provides the Configuration and ConfigBuilder classes.

Configuration class represents the immutable, read-only configuration object.

ConfigBuilder can be used to create new configuration from an existing one.
'''

from enum import Enum
from arjuna.tpi.constant import *
from arjuna.core.constant import *
import uuid
from arjuna.tpi.arjuna_types import *
from typing import *


class _ROWrapper:

    def __init__(self, config):
        self.__config = config

    def value(self, option):
        return self.__config.value(option)

    def items(self):
        return self.__config.items()

from arjuna.tpi.tracker import track

class Configuration:
    '''
        Read-only, immutable object that contains a fixed mapping of ArjunaOptions and User defined options.

        Args:
            test_session: Current test session object.
            name: Name of this configuration.
            config: WrappedConfiguration object. Configuration provides a read-only interface on top of it.

        Note:
            - You never directly create a Configuration object.
            - It is an outcome of configuration processing that Arjuna does.

        Note:
            - The option name string is considered by Arjuna as **case-insensitive**. Also, **. (dot)** and **_ (underscore)** are interchangeable. So, following are equivalent arguments
                - ArjunaOption.BROWSER_NAME
                - BROWSER_NAME
                - BrOwSeR_NaMe
                - browser.name
                - Browser.Name
                - and so on

        Note:
            You can also use **. notation** or **[] dict notation** to retrieve an option value.

            .. code-block:: python

                config.option
                config[option]
    '''

    def __init__(self, test_session: 'TestSession', name: str, config: 'WrappedConfiguration'):
        super().__init__()
        self.__session = test_session
        self.__name = name
        self.__wrapped_config = config
        self.__arjuna_options = _ROWrapper(self.__wrapped_config.arjuna_options)
        self.__user_options = _ROWrapper(self.__wrapped_config.user_options)

    @property
    def builder(self) -> 'ConfigBuilder':
        '''
            Creates a configuration builder object which takes this configuration as its reference.

            Returns:
                **new** `ConfigBuilder` object
        '''
        return ConfigBuilder(base_config=self)

    def value(self, option: ArjunaOptionOrStr) -> Any:
        '''
            Get the value of a configuration option.

            Args:
                option: An ArjunaOption or a string representing ArjunaOption or a user defined option.

            Returns:
                Object of any type, depending on the option.
        '''
        try:
            return self.__arjuna_options.value(option)
        except:
            try:
                return self.__user_options.value(option)
            except:
                raise Exception("No config option with name '{}' found in {} configuration.".format(option, self.name))

    @property
    def _wrapped_config(self):
        return self.__wrapped_config

    @property
    def test_session(self) -> 'TestSession':
        '''
            Test Session object for this Configuration.
        '''
        return self.__session

    def get_arjuna_options_as_map(self) -> Dict[ArjunaOptionOrStr, Any]:
        '''
            Get all Arjuna options.

            Returns:
                A dictionary of all Arjuna Options.
        '''
        return self.__wrapped_config.arjuna_options.as_dict()

    def is_arjuna_option_not_set(self, option) -> bool:
        '''
            Check if the value for an Arjuna option was set. (Checks for 'not_set' string.)

            Args:
                option: An ArjunaOption or a string representing ArjunaOption or a user defined option.

            Returns:
                True/False
        '''
        return self.__wrapped_config.arjuna_options.is_not_set(option)

    @property
    def name(self) -> str:
        '''
            Name of this configuration object.

            Returns:
                Name of this configuration
        '''
        return self.__name

    def __getattr__(self, name):
        return self.value(name)

    def __getitem__(self, name):
        return self.value(name)

    def __call__(self, name):
        return self.value(name)

    def as_dict(self) -> Dict[str, Any]:
        '''
            Get all options.

            Returns:
                A dictionary of all Arjuna Options and User Defined Options
        '''
        return self.__wrapped_config.as_dict()

    def __str__(self):
        return self.name

@track("trace")
class ConfigBuilder:
    '''
        Helps in constructing a new **Configuration** object from an existing one.

        Keyword Arguments:
            base_config: Parent configuration to be used be used as reference for creation of the new Configuration object.
            auto_gen_name: Generate a unique configuration name in **build** call if name is not provided. Default is True.

        Note:
            It is not meant to be directly constructed. Use **builder** method of a Configuration object to create the associated ConfigBuilder object.

        Note:
            You can also use **. notation** or **[] dict notation** to add/update an option value.

            .. code-block:: python

                builder.option = value
                builder[option] = value
    '''

    def __init__(self, *, base_config: Configuration, auto_name_gen=True, _conf_stage=ConfigStage.CODED):
        from arjuna.configure.options import EditableConfig
        vars(self)['_test_session'] = base_config.test_session
        vars(self)['_config_container'] = EditableConfig.empty_conf()
        vars(self)['_base_config'] = base_config
        vars(self)['_auto_gen_name'] = auto_name_gen
        vars(self)['_conf_stage'] = _conf_stage

    def option(self, option: ArjunaOptionOrStr, obj: Any) -> 'self':
        '''
            Add/Change option value.

            Args:
                option: An ArjunaOption or a string representing ArjunaOption or a user defined option.
                obj: An object of any type as per the option key.

            Returns:
                Current `ConfigBuilder` object
        '''
        self._config_container.set_option(option, obj)
        return self

    def __setattr__(self, option, obj) -> 'self':
        '''
            Enables **. notation** for setting option value. Same as **option** method.
        '''
        self.option(option, obj)
        return self

    def __setitem__(self, option, obj) -> 'self':
        '''
            Enables **[] dict notation** for setting option value. Same as **option** method.
        '''
        self.option(option, obj)
        return self

    def options(self, option_map: Dict[ArjunaOptionOrStr, Any]) -> 'self':
        '''
            Add/Change multiple option values.

            Args:
                option_map: A dictionary of options. Keys are ArjunaOptions or strings and values can be of any type as per the option key.

            Returns:
                Current `ConfigBuilder` object
        '''
        self._config_container.set_options(option_map)
        return self

    def selenium(self) -> 'self':
        '''
            Configure Selenium as the automation engine.

            Returns:
                Current `ConfigBuilder` object
        '''
        self.set_option(ArjunaOption.GUIAUTO_NAME, GuiAutomatorName.SELENIUM)
        return self

    def appium(self, context) -> 'self':
        '''
            (Not supported yet) Configure Appium as the automation engine.

            Returns:
                Current `ConfigBuilder` object
        '''
        self.option(ArjunaOption.GUIAUTO_NAME, GuiAutomatorName.APPIUM)
        self.option(ArjunaOption.GUIAUTO_CONTEXT, context)
        return self

    def chrome(self) -> 'self':
        '''
            Configure Chrome as the browser.

            Returns:
                Current `ConfigBuilder` object
        '''
        self.option(ArjunaOption.BROWSER_NAME, BrowserName.CHROME)
        return self

    def firefox(self) -> 'self':
        '''
            Configure Firefox as the browser.

            Returns:
                Current `ConfigBuilder` object
        '''
        self.option(ArjunaOption.BROWSER_NAME, BrowserName.FIREFOX)
        return self

    def app(self, path) -> 'self':
        '''
            (Not supported yet) Configure path for mobile app installer.

            Args:
                path: Absolute path of mobile app on test machine.

            Returns:
                Current `ConfigBuilder` object
        '''
        self.option(ArjunaOption.MOBILE_APP_FILE_PATH, path)
        return self

    def from_file(self, fpath) -> 'self':
        '''
            Add all options from a .conf file.

            Args:
                fpath: Path of .conf file. 

            Returns:
                Current `ConfigBuilder` object

            Note:
                If instead of full absolute path, a name or relative file path is provided, Arjuna creates the path in relation to the default configuration directory - **<Project Root>/config**.
        '''
        conf = self._test_session.load_options_from_file(fpath, conf_stage=self._conf_stage)
        for k,v in conf.arjuna_options.as_dict().items():
            self._config_container.set_arjuna_option(k,v)
        for k,v in conf.user_options.as_dict().items():
            self._config_container.set_user_option(k,v)

    def register(self, *, config_name=None):
        '''
            Register the new configuration.

            Keyword Arguments:
                config_name: (Optional) Name that you want to assign to the configuration.

            Note:
                - Name needs to be unique in a test run. No configuration should exist with the same name.
                - A dynamic, unique name is generated if name is not provided.

        '''
        from arjuna import Arjuna, ConfigCreationError
        if config_name is None and not self._auto_gen_name:
            raise ConfigCreationError("This ConfigBuilder has been created with auto_gen_name setting as False. You must provide a config_name to the build call.")
        config_name = config_name and config_name or 'c{}'.format(str(uuid.uuid4()).replace("-","_"))

        try:
            from arjuna.configure.validator import Validator
            Validator.name(config_name)
        except:
            raise ConfigCreationError("Unsupported name >>{}<< provided for a Configuration object. {}".format(config_name, Validator.VNREGEX_TEXT))

        if Arjuna.has_config(config_name):
            raise ConfigCreationError("You can not re-register a configuration for a name. Config with name {} already exists.".format(config_name))

        config = self._test_session.register_config(config_name.lower(), 
                                        self._config_container.arjuna_options, #.items(),
                                        self._config_container.user_options, #.items(),
                                        conf_stage=self._conf_stage,
                                        parent_config=self._base_config
                                    )
        return config


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
from arjuna.interact.gui.gom.namestore import GuiNameStore
from arjuna.tpi.constant import ArjunaOption

class GuiManager:

    def __init__(self, config):
        self.__name_store = GuiNameStore()
        self.__namespace_dir = config.arjuna_options.value(ArjunaOption.GUIAUTO_NAMESPACE_DIR)

    @property
    def name_store(self):
        return self.__name_store

    @property
    def namespace_dir(self):
        return self.__namespace_dir

    def define_gui(self, automator, label=None, name=None, qual_name=None, def_file_path=None):
        from arjuna.interact.gui.gom.guidef import GuiDef
        gui_def = GuiDef(self.__name_store, self.__namespace_dir, automator, label, def_file_path)
        gui = Gui(self, automator, gui_def)
        return gui





# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Data Generator

This module contains classes and functions that generate data.
'''

import uuid
import random
from mimesis import Person
from mimesis import Address
from mimesis import locales
from mimesis import Text

Locales = locales

class _gen:
    pass

class processor:
    '''
        Processes data by calling the provided callable with provided arguments.

        Arguments:
            callable: The callable to be called in processing.
            *args: Arbitrary positional args to be passed to the callable.

        Keyword Arguments:
            **kwargs: Arbitrary keyword arguments to tbe passed to the callable.
    '''

    def __init__(self, callable, *args, **kwargs):
        self.__callble = callable
        self.__args = args
        self.__kwargs = kwargs

    def process(self, data):
        '''
            Processes data in the provided data object by calling the callable with arguments provided in constructor.

            Arguments:
                data: Data object
        '''
        if type(self.__callble) is str:
            return getattr(data, self.__callble)(*self.__args, **self.__kwargs)
        else:
            return self.__callble(data_iterator, *self.__args, **self.__kwargs)  

class composer:
    '''
        Combines data in an iterable by calling the provided callable with provided arguments.

        Arguments:
            callable: The callable to be called in composing.
            *args: Arbitrary positional args to be passed to the callable.

        Keyword Arguments:
            **kwargs: Arbitrary keyword arguments to tbe passed to the callable.
    '''

    def __init__(self, callable, *args, **kwargs):
        self.__callble = callable
        self.__args = args
        self.__kwargs = kwargs

    def compose(self, data_iter):
        '''
            Combine data in the provided iterable by calling the callable with arguments provided in constructor.

            Arguments:
                data_iter: Data iterable
        '''
        return self.__callble(data_iter, *self.__args, **self.__kwargs)  

def _same(in_data):
    return in_data

class generator(_gen):
    '''
        Generate data by calling the provided callable with provided arguments.

        Arguments:
            callable: The callable to be called in composing. By default, same generated object is returned.
            *args: Arbitrary positional args to be passed to the callable.

        Keyword Arguments:
            processor: This callable is called after data is generated by passing generated data as argument. Useful for data transformation of any kind. If its type is string, it is assumed to be a method of the data which the data callable generated.
            **kwargs: Arbitrary keyword arguments to tbe passed to the callable.
    '''

    def __init__(self, callable, *args, processor=_same, **kwargs):
        self.__callble = callable
        self.__args = args
        self.__kwargs = kwargs
        self.__processor = processor

    def generate(self):
        '''
            Generate data by calling the callable with arguments provided in constructor.

            After generation the coverter callable is called with this data before returning the data.
        '''
        data = self.__callble(*self.__args, **self.__kwargs)
        if type(self.__processor) is str:
            return getattr(data, self.__processor)()
        elif isinstance(self.__processor, processor):
            return self.__processor.process(data)
        else:
            return self.__processor(data)

class composite(_gen):
    '''
        Composite Data Generator.

        Generate data by composing the output of all generators, callables or static data.

        Arguments:
            *generators_or_data: Arbitrary generators, functions or static data objects.

        Keyword Arguments:
            composer: This callable is called after data sequence is generated by passing generated data sequence as argument. Useful for data transformation of any kind.
    '''

    def __init__(self, *generators_or_data, composer=_same):
        self.__generators_or_data = generators_or_data
        self.__composer = composer

    def generate(self):
        '''
            Generate data by composing the output of all generators, callables or static data.
            
            The composer callable is called after data sequence is generated by passing generated data sequence as argument. Useful for data transformation of any kind.
        '''
        out = list()
        for gen_or_data in self.__generators_or_data:
            if hasattr(gen_or_data, '__call__'):
                out.append(gen_or_data())
            elif isinstance(gen_or_data, generator):
                out.append(gen_or_data.generate())
            else:
                out.append(gen_or_data)

        if isinstance(self.__composer, composer):
            return self.__composer.compose(out)
        else:
            return self.__composer(out)

class Random:
    '''
        Provides methods to create random strings and numbers of different kinds.
    '''

    @classmethod
    def ustr(cls, *, prefix: str=None, maxlen=None) -> str:
        '''
            Generate a unique UUID string

            Keyword Arguments:
                prefix: (Optional) prefix to be added to the generated UUID string.
                maxlen: (Optional) Maximum length of the retuned string (inclusive of prefix-).

            Returns:
                A string that is unique for current session.
        '''
        prefix = prefix and prefix + "-" or ""
        return maxlen and "{}{}".format(prefix, uuid.uuid4())[:int(maxlen)] or "{}{}".format(prefix, uuid.uuid4())

    @classmethod
    def first_name(cls, *, locale=Locales.EN):
        '''
            Generate a first name.

            Keyword Arguments:
                locale: (Optional) locale for generating first name

            Returns:
                A generated first name
        '''
        return Person(locale).first_name()

    @classmethod
    def last_name(cls, *, locale=Locales.EN):
        '''
            Generate a last name.

            Keyword Arguments:
                locale: (Optional) locale for generating last name

            Returns:
                A generated last name
        '''
        return Person(locale).last_name()

    @classmethod
    def name(cls, *, locale=Locales.EN):
        '''
            Generate a full name (first name and last name).

            Keyword Arguments:
                locale: (Optional) locale for generating phone number

            Returns:
                A generated full name
        '''
        return "{} {}".format(cls.first_name(locale=locale), cls.last_name(locale=locale))

    @classmethod
    def phone(cls, *, locale=Locales.EN):
        '''
            Generate a phone number.

            Keyword Arguments:
                locale: (Optional) locale for generating phone number

            Returns:
                A generated phone number
        '''
        return cls.first_name(locale=locale)

    @classmethod
    def email(cls, *, locale=Locales.EN):
        '''
            Generate an email address.

            Keyword Arguments:
                locale: (Optional) locale for generating email address

            Returns:
                A generated email address
        '''
        return Person(locale).email()

    @classmethod
    def street_name(cls, *, locale=Locales.EN):
        '''
            Generate a street name

            Keyword Arguments:
                locale: (Optional) locale for generating street name

            Returns:
                A generated street name
        '''
        return Address(locale).street_name()

    @classmethod
    def street_number(cls, *, locale=Locales.EN):
        '''
            Generate a street number

            Keyword Arguments:
                locale: (Optional) locale for generating street number

            Returns:
                A generated street number
        '''
        return Address(locale).street_number()

    @classmethod
    def house_number(cls, *, locale=Locales.EN):
        '''
            Generate a house number

            Keyword Arguments:
                locale: (Optional) locale for generating house number

            Returns:
                A generated house number
        '''
        return cls.street_number(locale=locale)

    @classmethod
    def postal_code(cls, *, locale=Locales.EN):
        '''
            Generate a postal code

            Keyword Arguments:
                locale: (Optional) locale for generating postal code

            Returns:
                A generated postal code
        '''
        return Address(locale).postal_code()

    @classmethod
    def city(cls, *, locale=Locales.EN):
        '''
            Generate a city name.

            Keyword Arguments:
                locale: (Optional) locale for generating city name

            Returns:
                A generated city name
        '''
        return Address(locale).city()

    @classmethod
    def country(cls, *, locale=Locales.EN):
        '''
            Generate a country name

            Keyword Arguments:
                locale: (Optional) locale for generating country name

            Returns:
                A generated country name
        '''
        return Address(locale).country()

    @classmethod
    def sentence(cls, *, locale=Locales.EN):
        '''
            Generate a sentence

            Keyword Arguments:
                locale: (Optional) locale for generating sentence

            Returns:
                A generated sentence
        '''
        return Text(locale).sentence()

    @classmethod
    def fixed_length_number(cls, *, length):
        '''
            Generate a fixed length number

            Keyword Arguments:
                length: Number of digits in generated number.

            Returns:
                A generated fixed length number

            Note:
                A number of minimum length 1 is always generated.
        '''
        arr0 = [str(random.randint(1,9))]
        arr = []
        if length >= 2:
            arr = [str(random.randint(0,9)) for i in range(length-1)]
        arr0.extend(arr)
        return int("".join(arr0))

    @classmethod
    def int(cls, *, end, begin=0):
        '''
            Generate a random integer.

            Keyword Arguments:
                end: (inclusive) upper limit for the integer
                begin: (inclusive) lower limit for the integer. Default is 0.

            Returns:
                A generated integer
        '''
        return random.randint(begin, end)

        



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Log module contains helper functions for easy and pwoerful logging with Arjuna.

All log messages are directed to console as well as arjuna.log file which is generated on a per run basis and is placed in **<project_root>/report/run_dir/log** directory.

Messages generated by the test project as well as Arjuna go to the same log for easy checking of flow.

6 levels of logging are available:

    #. TRACE
    #. DEBUG
    #. INFO
    #. WARNING
    #. ERROR
    #. FATAL

These are represented by six different log_* functions.

**TRACE** has the lowest priority and **FATAL** has the highest priority amongst message levels.

Following Arjuna Options are related to logging:
    * LOG_CONSOLE_LEVEL: Minimum level of logging for a run for displaying log messages on console.
    * LOG_CONSOLE_LEVEL: Minimum level of logging for a run for displaying log messages in arjuna.log.
    * LOG_ALLOWED_CONTEXTS: The context strings which determine log messages belonging to which contexts can be displayed and logged.
'''

import sys

from arjuna.tpi.helper.audit import _Stack
from arjuna.tpi.arjuna_types import *

def __log(invoker, level, msg, contexts=None):
    from arjuna import Arjuna, ArjunaOption
    if type(contexts) is str:
        contexts = (contexts,)
    elif contexts is None:
        contexts = ("default",)
    contexts = set(contexts)
    try:
        getattr(Arjuna.get_logger(), level)(msg.replace('\n', ' ').replace('\r', ''), extra={'invoker': invoker, 'contexts':contexts})
    except AttributeError:
        # In case the logging is called before the logger is set.
        # In future versions, see if there can be a fallabck logger.
        pass
    except OSError:
        ## On Windows 10, random handle related bugs happen.
        if level.lower() in {"info", "debug", "trace"} :
            sys.stdout.write(msg+ "\n")
        else :
            sys.stderr.write(msg+"\n")

def log_trace(msg: str, *, contexts: ListOrTupleOrStr=None):
    '''
        Log a message with **TRACE** level.

        Args: 
            msg: Log message
            contexts: (Optional) Context strings for this log message.
    '''
    __log(_Stack.get_invoker(), "trace", msg, contexts=contexts)

def log_debug(msg: str, *, contexts: ListOrTupleOrStr=None) -> None:
    '''
        Log a message with **DEBUG** level.

        Args: 
            msg: Log message
            contexts: (Optional) Context strings for this log message.
    '''
    __log(_Stack.get_invoker(), "debug", msg, contexts=contexts)

def log_info(msg: str, *, contexts: ListOrTupleOrStr=None) -> None:
    '''
        Log a message with **INFO** level.

        Args: 
            msg: Log message
            contexts: (Optional) Context strings for this log message.
    '''
    __log(_Stack.get_invoker(), "info", msg, contexts=contexts)

def log_warning(msg: str, *, contexts: ListOrTupleOrStr=None) -> None:
    '''
        Log a message with **WARNING** level.

        Args: 
            msg: Log message
            contexts: (Optional) Context strings for this log message.
    '''
    __log(_Stack.get_invoker(), "warning", msg, contexts=contexts)

def log_error(msg: str, *, contexts: ListOrTupleOrStr=None) -> None:
    '''
        Log a message with **ERROR** level.

        Args: 
            msg: Log message
            contexts: (Optional) Context strings for this log message.
    '''
    __log(_Stack.get_invoker(), "error", msg, contexts=contexts)

def log_fatal(msg: str, *, contexts: ListOrTupleOrStr=None) -> None:
    '''
        Log a message with **FATAL** level.

        Args: 
            msg: Log message
            contexts: (Optional) Context strings for this log message.
    '''
    __log(_Stack.get_invoker(), "fatal", msg, contexts=contexts)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.core.poller.caller import DynamicCaller
from arjuna.core.poller.conditions import CommandCondition

class GuiGuiWidgetContainerConditions:

    def __init__(self, container):
        self.__container = container

    @property
    def container(self):
        return self.__container

    def PresenceOfElement(self, gui_element):
        caller = DynamicCaller(
            self.container._element_finder.find,  
            self.__container.dispatcher.find_element,
            gui_element._wmd,
            context = "ELEMENT"
        )
        return CommandCondition(caller)   

    def PresenceOfMultiElement(self, gui_element):
        caller = DynamicCaller(
            self.container._element_finder.find, 
            self.container.dispatcher.find_multielement,
            gui_element._wmd,
            context = "MULTI_ELEMENT"
        )
        return CommandCondition(caller)  

    def AbsenceOfElement(self, wmd):
        caller = DynamicCaller(
            self.container._element_finder.check_for_absence, 
            self.container.dispatcher.find_element,
            wmd,
            context = "ELEMENT"
        )
        return CommandCondition(caller) 


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import abc
from urllib.parse import urlparse, urlencode, parse_qs, quote
from requests import Request, Session
from arjuna.tpi.error import HttpUnexpectedStatusCodeError, HttpSendError
from arjuna.tpi.parser.json import Json
from arjuna.tpi.parser.html import Html
from arjuna.tpi.engine.asserter import AsserterMixIn
from arjuna.tpi.helper.arjtype import CIStringDict
from requests.exceptions import ConnectionError, TooManyRedirects
import time

from .packet import HttpPacket


class HttpResponse(HttpPacket):
    '''
        Encapsulates HTTP response packet. Contains redirected responses as redirection history, if applicable.

        Arguments:
            session: `HttpSession` object which created corresponding `HttpRequest` for this response.
            response: `requests` library's Response object wrapped by this class.
    '''

    def __init__(self, session, response):
        super().__init__(response)
        self.__session = session
        self.__resp = response
        self.__store = CIStringDict()

    @property
    def is_redirect(self):
        '''
        Is True if this is a response is a redirection response.
        '''
        return self.__resp.is_redirect

    @property
    def store(self):
        '''
        Values extracted from response and stored.
        '''
        return self.__store

    @property
    def url(self) -> str:
        ''' 
            URL for which this response was generated.

            In case of redirections, this is the last URL requested.
        '''
        return self.__resp.url

    @property
    def query_params(self) -> dict:
        ''' 
            Query parameters in URL for this response.

            In case of redirections, these are the query parameters in last request.
        '''
        return parse_qs(self.__resp.url)

    @property
    def status_code(self) -> int:
        ''' 
            HTTP Status code for this response. For example, 200
        '''
        return self.__resp.status_code

    def assert_status_codes(self, codes, *, msg):
        '''
            Assert that the status code is as expected.

            Arguments:
                codes: str or iterator

            Keyword Arguments:
                msg: Purpose of this assertion
        '''
        if type(codes) is int:
            codes = {codes}
        self.asserter.assert_true(self.status_code in codes, f"HTTP status code {self.status_code} is not expected. Expected: {codes}. {msg}")

    def assert_headers(self, headers: dict, *, msg):
        '''
            Assert that all headers in provided in the headers dict are matched in response.

            Arguments:
                headers: HTTP headers dictionary

            Keyword Arguments:
                msg: Purpose of this assertion
        '''
        for k,v in headers.items():
            self.asserter.assert_true(k in self.headers, f"HTTP response header {k} is not present in response. {msg}")
            self.asserter.assert_equal(self.headers[k], v, f"HTTP response header {k} value does not match. {msg}")

    @property
    def status(self) -> str:
        ''' 
            HTTP Status Message for this response. For example, Not Found
        '''
        return self.__resp.reason

    @property
    def text(self):
        '''
            HTTP Response content as Text object.
        '''
        from arjuna.tpi.parser.text import Text
        return Text(self.content)

    @property
    def content(self):
        '''
            HTTP Response content as plain unformatted text.
        '''
        return self.__resp.text

    @property
    def json(self) -> 'JsonDictOrJsonList':
        ''' 
            HTTP Response content as Arjuna's `JsonDict` or `JsonList` object.
        '''
        return Json.from_str(self.content)

    @property
    def html(self) -> 'HtmlNode':
        ''' 
            HTTP Response content as Arjuna's `HtmlNode` object.
        '''
        return Html.from_str(self.content)

    @property
    def redir_history(self) -> tuple:
        '''
            Ordered `HttpResponse` objects for all redirections that led to this response.
        '''
        if self.__resp.history:
            return tuple([HttpResponse(self.__session, h) for h in self.__resp.history])
        else:
            return tuple()

    @property
    def last_redir_response(self) -> 'HttpResponse or None':
        '''
            Last `HttpResponse` object in case of redirections. None in case of no reidrections.
        '''
        if not self.redir_history:
            return None
        return self.redir_history[-1]

    @property
    def next_request(self):
        '''
            Next `HttpRequest` object if this response redirects to another request. None in case this is the last response in chain.
        '''
        from .request import HttpRequest
        next_req = self.__resp.next
        if next_req:
            return HttpRequest(self.__session, self.__resp.next)
        else:
            return None

    @property
    def request(self):
        '''
            `HttpRequest` object corresponding to this response object.
        '''
        from .request import HttpRequest
        return HttpRequest(self.__session, self.__resp.request)   

    __OUT = '''{}
{}{}'''

    @classmethod
    def __try_as_json(self, text):
        if text is None: return ""
        try:
            return json.dumps(json.loads(text), indent=2)
        except:
            return text

    @classmethod
    def repr_as_str(cls, *, status_code, status_msg, headers, content=None):
        content = cls.__try_as_json(content)
        if content:
            content = '\n\n{}\n'.format(content)
        return cls.__OUT.format(
            str(status_code) + ' ' + status_msg,
            '\n'.join('{}: {}'.format(k, v) for k, v in headers.items()),
            content
        ).strip()

    def __str__(self):
        return self.repr_as_str(
            status_code = self.status_code,
            status_msg = self.status,
            headers = self.headers,
            content = self.content
        )




# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.engine.asserter import AsserterMixIn

class CookieValidator(AsserterMixIn):

    def __init__(self, session, cookie_yaml):
        super().__init__()
        cookies = session.parsed_cookies
        for cookie_name, cookie_val in cookie_yaml.items():
            msg = f"Cookie with name >>{cookie_name}<< was not found in session cookies."
            try:
                self.asserter.assert_true(cookie_name in session.cookies, msg=msg)
            except AssertionError:
                self.asserter.fail(msg=msg)

            cookie = cookies[cookie_name]

            if type(cookie_val) is dict:
                for k,v in cookie_val.items():
                    if k == "value":
                        self.asserter.assert_equal(cookie.value, cookie_val["value"], f"Cookie value for >>{cookie_name}<< does not match expected value.")
                    elif k.lower() == "httponly":
                        self.asserter.assert_true(cookie.httponly, f"Cookie >>{cookie_name}<< does not have >>HttpOnly flag<< set.")
                    elif k.lower() == "secure":
                        self.asserter.assert_true(cookie.secure, f"Cookie >>{cookie_name}<< does not have >>secure flag<< set.")
            else:
                self.asserter.assert_equal(cookie.value, str(cookie_val), f"Cookie value for >>{cookie_name}<< does not match expected value.")



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import abc


class Database(metaclass=abc.ABCMeta):

    def __init__(self, db_handle, config):
        self.__handle = db_handle
        from arjuna import Arjuna
        self.__config = config is not None and config or Arjuna.get_config()

    @property
    def _handle(self):
        return self.__handle

    @property
    def _config(self):
        return self.__config

    def execute_all(self, *queries, merge=True, **formatters):
        cursor = self._handle.cursor()
        if merge:
            queries = ";".join(queries).format(**formatters)
            print(queries)
            cursor.execute(queries, multi=True)
        else:
            for query in queries:
                query = query.format(**formatters)
                print(query)
                cursor.execute(query)

        self._handle.commit()
        cursor.close()

    def __get_sql_file_path(self, fname):
        from arjuna.core.utils import file_utils
        from arjuna import Arjuna, ArjunaOption
        fpath = os.path.abspath(os.path.join(self._config.value(ArjunaOption.DBAUTO_SQL_DIR), fname))
        if file_utils.is_file(fpath):
            return fpath
        else:
            for linked_project in reversed(Arjuna.get_linked_projects()):
                fpath = os.path.abspath(os.path.join(linked_project.ref_conf.value(ArjunaOption.DBAUTO_SQL_DIR), fname))
                if file_utils.is_file(fpath):
                    return fpath
        raise Exception("DBAuto SQL File not found: {}".format(fname))

    def execute_file(self, fpath, **formatters):
        fpath = self.__get_sql_file_path(fpath)
        from arjuna.tpi.parser.text import TextFileAsLines
        reader = TextFileAsLines(fpath)
        queries = list()
        for query in reader:
            if query != "":
                queries.append(query)
        return self.execute_all(*queries, **formatters)

    def close(self):
        self._handle.close()


class MySQL(Database):

    def __init__(self, db_handle, config):
        super().__init__(db_handle, config)


class DB:

    @classmethod
    def mysql(cls, *, host, user, password, port=3306, config=None):
        import mysql.connector
        db = mysql.connector.connect(
            host=host,
            user=user,
            password=password,
            ssl_disabled = True,

        )
        return MySQL(db, config)





# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser


class IniFile2MapReader:
    def __init__(self, file_path):
        self.file_path = file_path
        self._f = configparser.ConfigParser()
        self._f.read(self.file_path)
        self.sections = self._f.sections()

    def __get_section_data(self, section):
        return self._f[section]

    def __iter__(self):
        return self

    def next(self):
        if self.sections:
            return self.__get_section_data(self.sections.pop(0))
        else:
            raise StopIteration()

    def read(self):
        section_data = {}
        for section in self.sections:
            section_data[section] = dict(self._f[section])
        return section_data

    def close(self):
        self._f.clear()
        del self._f




from arjuna.tpi.guiauto.meta.formatter import GNSLabelFormatter
from arjuna.tpi.error import *
from arjuna.core.error import *
from arjuna.interact.gui.auto.finder import GuiEmdFinder, GuiElementEmdFinder

from arjuna import track

@track("debug")
class GNS:
    '''
        Gui Namespace object associated with a **Gui** or **GuiElement**.

        Not meant to be directly created. Available as **gns** property of a **Gui**.

        Arguments:
            gui_or_element: **Gui** or **GuiElement** object.
            gui_def: Gui Definition

        Note:
            To locate a GUI Widget using GNS object, you can use the **.** notation. For example:

                .. code-block:: python

                    gui.gns.some_label
                    some_element.gns.some_label

            This will locate the GuiWidget using the GuiWidgetMetaData corresponding to following entry in GNS file:

                .. code-block:: YAML

                    labels:
                        some_label:
                            <metdata dictionary>

            **gui.gns** will locate it in the root GuiWidget of **Gui** if available and allowed.

            **some_element.gns.some_label** will do a nested locating inside **some_element**.
    '''

    def __init__(self, gui_or_element, gui_def):
        self.__container_type = None
        self.__gui = None
        self.__container = None
        self.__process_container(gui_or_element)
        self.__gui_def = gui_def
        if self.__container_type == "gui":
            self.__finder = GuiEmdFinder(self.__gui)
        else:
            self.__finder = GuiElementEmdFinder(self.__container)

    def _get_gui_def(self):
        return self.__gui_def

    def __process_container(self, gui_or_element):
        from .gui import Gui
        if isinstance(gui_or_element, Gui):
            if not gui_or_element._root_element:
                self.__container_type = "gui"
                self.__container = gui_or_element
                self.__gui = gui_or_element
            else:
                self.__container_type = "element"
                self.__container = gui_or_element._root_element
                self.__gui = gui_or_element
        else:
            self.__container_type = "element"
            self.__container = gui_or_element 
            self.__gui = gui_or_element.gui  
        self.__loaded = True         

    def formatter(self, **fargs) -> GNSLabelFormatter:
        '''
            Create a :class:`~arjuna.tpi.guiauto.meta.formatter.GNSLabelFormatter` object.

            Keyword Arguments:
                **fargs: Arbitrary key-value pairs to be used for formatting identifiers in **GuiWidgetDefinition**.
        '''
        return GNSLabelFormatter(self, **fargs)

    def _get_wmd_for_label(self, label):
        wmd = None
        try:
            wmd = self.__gui_def.get_wmd(label)
        except Exception as e:
            if self.__gui.app._common_guidef is not self.__gui_def:
                wmd = self.__gui.app._common_guidef.get_wmd(label)
        if wmd is None:
            raise Exception(f"Definition for label {label} not found.")
        return wmd.create_formatted_wmd() # Only globals will be processed.

    def wait_until_absent(self, *labels):
        '''
            Wait until **GuiWidget** corresponding to any of the GNS labels is present.

            Args:
                *labels: One or more GNS labels.

            Note:
                By default Wait is done until **ArjunaOption.GUIAUTO_MAX_WAIT** in the **Configuration** object associated with this **GuiAppContent**.

                You can provide **max_wait** key-value for GNS label(s) in GNS file to change this. Value is considered in seconds.
        '''
        waiter = getattr(self.__container, "_" + "wait_until_absent")
        for label in labels:
            wmd = self._get_wmd_for_label(label)
            try:
                waiter(wmd)
            except GuiWidgetPresentError:
                raise GuiWidgetForLabelPresentError(self.__gui, label)    

    def contains(self, *labels):
        '''
            Check whether this **GuiAppContent** object contains a **GuiWidget** corresponding to any of the GNS labels. Includes dynamic waiting.

            Keyword Arguments:
                *labels: One or more GNS labels.

            Note:
                By default Wait is done until **ArjunaOption.GUIAUTO_MAX_WAIT** in the **Configuration** object associated with this **GuiAppContent**.

                You can provide **max_wait** key-value for GNS label(s) in GNS file to change this. Value is considered in seconds.
        '''
        for label in labels:
            try:
                getattr(self, label)
            except GuiWidgetForLabelNotPresentError:
                continue
            else:
                return True
        return False

    def __process_labels_in_relations(self, wmd):
        from arjuna import log_debug
        log_debug("Processing relationship for GuiWidgetMetaData: {} in gui: {}".format(wmd, self.__gui))
        for k,v in wmd.meta.relations.items():
            from arjuna.interact.gui.auto.finder.wmd import GuiWidgetMetaData
            if type(v) is str:
                log_debug("Triggering locating operation for label {} in relations dict.".format(k))
                wmd.meta.relations[k] = getattr(self, v).dispatcher.driver_element
                log_debug("Replaced label {} with corresponding GuiElement in relations dict.".format(k))               
        return wmd

    def _locate_with_wmd(self, wmd):
        wmd = self.__process_labels_in_relations(wmd)
        factory = getattr(self.__finder, wmd.meta["type"].name.lower())
        return factory(wmd)

    def __getattr__(self, label):
        wmd = self._get_wmd_for_label(label)
        from arjuna import log_debug
        log_debug("Finding element with label: {} and wmd: {}".format(label, wmd))
        try:
            return self._locate_with_wmd(wmd)
        except ArjunaTimeoutError:
            raise GuiWidgetForLabelNotPresentError(self.__gui, label)



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.tracker import track

from .base import GuiSource
from .content import GuiSourceContent

@track("trace")
class _GuiFrameSource(GuiSource):
    '''
        Not supported yet.
    '''

    def __init__(self, frame):
        super().__init__()
        self.__frame = frame
        self.__root_source = None
        self.__html_source = None

    def set_root_source(self, src):
        '''
        Once frame switch takes place, this source can not be got. Hence needs to happen
        explicitly at the time of wrapped element finding.
        '''
        self.__root_source = src

    def _load(self):
        self.__frame.focus()
        self.__html_source = self.__frame._get_html_content_from_remote()

    def get_full_content(self):
        return self.get_root_content() + self.get_inner_content()

    def get_inner_content(self):
        return self.__html_source.get_full_content()

    def get_text_content(self):
        return self.__html_source.get_text_content()

    def get_root_content(self):
        return self.__root_source


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.helper.arjtype import Dictable
from arjuna.tpi.tracker import track

@track("trace")
class GuiWidgetDefinition(Dictable):
    '''
        Representation of indentifiers, format arguments, widget configuration and named arguments for a GuiWidget.

        Keyword Arguments:
            type: type of GuiWidget. Default is 'element'/GuiWidgetType.ELEMENT. 
            fmt_args: Dictionary of key-value pairs to format the identifiers.
            **named_args: Named arguments representing allowed identifiers, extended identifiers, configuration options and GuiWidget arguments.
    '''

    def __init__(self, *, type="element", fmt_args=None, **named_args):
        self.__widget_type = type

        self.__fmt_args = fmt_args
        if self.__fmt_args is None:
            self.__fmt_args = dict()

        self.__named_args = named_args

        from arjuna.interact.gui.auto.finder.meta import Meta
        self.__meta = Meta({
            "type" : type,
        })

    @property
    def widget_type(self):
        '''
            Type of this GuiWidget.
        '''
        return self.__widget_type

    @property
    def fmt_args(self):
        '''
            Dictionary of key-value pairs to format the identifiers.
        '''
        return self.__fmt_args

    @property
    def named_args(self):
        '''
            Named arguments representing allowed identifiers, extended identifiers, configuration options and GuiWidget arguments.
        '''
        return self.__named_args

    def __str__(self):
        return str(self.as_dict())

    def _as_dict(self):
        return {
                "type": self.widget_type,
                "fmt_args": self.fmt_args,
                "meta": str(self.__meta)
            }

    def _as_raw_wmd(self):
        from arjuna import Arjuna, oneof, log_debug
        from arjuna.interact.gui.auto.finder._with import With
        from arjuna.interact.gui.auto.finder.enums import WithType

        def add_to_list(with_list, k, v):
            if k.upper() in WithType.__members__:
                if isinstance(v, Dictable):
                    v = v.as_dict()
                with_list.append(getattr(With, k.lower())(v))
            elif Arjuna.get_withx_ref().has_locator(k):
                if isinstance(v, Dictable):
                    v = v.as_dict()
                    with_list.append(getattr(With, k.lower())(**v))
                elif type(v) in {list, tuple}:
                    with_list.append(getattr(With, k.lower())(*v))
                else:
                    with_list.append(getattr(With, k.lower())(v))
            else:
                self.__meta[k] = v

        with_list = []
        for k,v in self.__named_args.items():
            if isinstance(v, oneof):
                for oneofval in v.as_list():
                    add_to_list(with_list, k, oneofval)
            else:
                add_to_list(with_list, k, v)

        if not with_list:
            from arjuna.tpi.error import GuiWidgetDefinitionError
            raise GuiWidgetDefinitionError("You must provide atleast one valid locator argument to create GuiWidgetDefinition definition. None of Arjuna defined locators or withx locators were provided. Got the speification dictionary as: {}".format(self.__named_args))
        from arjuna.interact.gui.auto.finder.wmd import GuiWidgetMetaData

        return GuiWidgetMetaData.create_wmd(*with_list, meta=self.__meta)

    def _as_wmd(self):
        '''
            Convert this **GuiWidgetDefinition** to **GuiWidgetMetaData** object
        '''
        wmd = self._as_raw_wmd()
        from arjuna import log_debug
        log_debug("Unformatted Widget def is: " + str(wmd))
        fmt_wmd = wmd.create_formatted_wmd(**self.__fmt_args)
        return fmt_wmd


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast

safe_eval = ast.literal_eval

def convert_yaml_obj_to_content(jval):
    def handle_str_val(in_val):
        if in_val.lower().strip() == "null":
            out_val = "null"
        else:
            # The string could be a variable name. With spaces, it will be treated as Python statement.
            if in_val in locals() or in_val in globals():
                out_val = in_val
            else:
                try:
                    out_val = safe_eval(in_val)
                except (NameError, SyntaxError, ValueError): # The string is evaluated as a variable name or Python statement
                    out_val = jval
        return out_val

    if type(jval) is dict:
        # This is needed if e.g. "1" is a string and not to be treated as a number. In this case type: str should be coded.
        if "value" in jval:
            raw_value = jval["value"]
            if "type" in jval:
                jtype = jval["type"].lower().strip()
                allowed_set = {"str", "dict", "list", "int", "float", "bool"}
                if jtype in allowed_set:
                    jval = eval(jtype)(jval["value"])
                else:
                    raise Exception("type can be specified only as one of {}".format(allowed_set))
            elif type(raw_value) is str:
                jval = handle_str_val(raw_value)
            else:
                jval = raw_value
        else:
            jval = jval
    elif type(jval) is str:
        jval = handle_str_val(jval)
    else:
        jval = jval

    return jval


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import random
from enum import Enum, auto

from arjuna.tpi.tracker import track

class pos:
    '''
        Factory Class with various factory methods to create objects for a single position and multiple positions.
    '''

    @classmethod
    def _create_extractor(cls, etype, *vargs, **kwargs):
        return getattr(cls, etype.lower())(*vargs, **kwargs)

    @classmethod
    def at(cls, *pos, strict: bool=True):
        '''
            Create Extractor object to extract objects at one or more positions from tuple/list.

            Args:
                pos: (int) Arbitrary positions in a sequence

            Keyword Arguments:
                strict: If True, exception is raised if no element is found at provided position.
        ''' 
        if len(pos) == 1:
            return _AtPositionExtractor(pos[0], strict=strict)
        else:
            return _AtPositionsExtractor(*pos, strict=strict)

    @classmethod
    def first(cls, strict: bool=True):
        '''
            Create Extractor object to extract objects at first position from tuple/list.

            Keyword Arguments:
                strict: If True, exception is raised if no element is found at provided position.
        ''' 
        return _FirstPositionExtractor(strict=strict)

    @classmethod
    def last(cls, strict: bool=True):
        '''
            Create Extractor object to extract objects at last position from tuple/list.

            Keyword Arguments:
                strict: If True, exception is raised if no element is found at provided position.
        ''' 
        return _LastPositionExtractor(strict=strict)


    @classmethod
    def random(cls, count=1, strict: bool=True):
        '''
            Create Extractor object to extract objects at random position(s) fom tuple/list.

            Arguments:
                count: Number of unique random positions to be used.
            
            Keyword Arguments:
                strict: If True, exception is raised if input sequence is empty or number of random samples requested > length of sequence.
        ''' 
        count = int(count)
        if count == 1:
            return _RandomPositionExtractor(strict=strict)
        else:
            return _RandomPositionsExtractor(count=count, strict=strict)

    @classmethod
    def slice(cls, *vargs, strict: bool=True, **kwargs):
        '''
            Create Extractor object to extract a slice of the provided tuple/list.

            Args:
                *vargs: (int) 1,2,or 3 arguments are accepted. See Notes.

            Keyword Arguments:
                kwargs: One or more of start/stop/step passed as keyword argument.
                strict: If True, exception is raised if extracted sequence is empty.

            Note:
                Following is the meaning of 1,2,3 arg signatures:

                .. code-block:: python

                    slice(stop)
                    slice(start, stop)
                    slice(start, stop, step)

                You can also use the keyword arguments instead of positional arguments:

                .. code-block:: python

                    slice(stop=4)
                    slice(start=2, stop=4)
                    slice(start=2, stop=7, step=2)

                You can not mix these styles of calling. Either use vargs or kwargs, else an Exception is raised.
        ''' 
        if vargs and kwargs:
            raise Exception("You can either use positional args or keword args to create a slice. These styles can not be mixed.")
        if not vargs and not kwargs:
            return _SlicedPositionsExtractor(strict=strict)
        if vargs:
            if len(vargs) > 3:
                raise Exception(f"Invalid positional arguments provided. Refer doc for 1,2,3 positional arg signature of this method. Args provided: {vargs}")
            
            vargs = [int(v) for v in vargs]
            if len(vargs) == 1:
                return _SlicedPositionsExtractor(stop=vargs[0], strict=strict)
            elif len(vargs) == 2:
                return _SlicedPositionsExtractor(start=vargs[0], stop=vargs[1], strict=strict)
            else:
                return _SlicedPositionsExtractor(start=vargs[0], stop=vargs[1], step=vargs[2], strict=strict)

        if kwargs:
            for k in kwargs:
                if k.lower() not in {"start", "stop", "step"}:
                    raise Exception(f"Invalid kwyword arguments provided. Only start/stop/step args are supported. Args provided: {kwargs}")
            kwargs = {k: k in {"start", "stop", "step"} and int(v) or v for k,v in kwargs.items()}
            return _SlicedPositionsExtractor(strict=strict, **kwargs)
                    

    @classmethod
    def odd(cls, strict: bool=True):
        '''
            Create Extractor object to extract objects at odd positions from given tuple/list.

            Keyword Arguments:
                strict: If True, exception is raised if extracted sequence is empty.
        ''' 
        return _OddPositionsExtractor(strict=strict)


    @classmethod
    def even(cls, strict: bool=True):
        '''
            Create Extractor object to extract objects at even positions from given tuple/list.

            Keyword Arguments:
                strict: If True, exception is raised if extracted sequence is empty.
        ''' 
        return _EvenPositionsExtractor(strict=strict)

@track("trace")
class Extractor(metaclass=abc.ABCMeta):
    '''
        Abstract class for extractors. Given an object, a concrete implementation extracts objects from the given object.

        Keyword Arguments:
            strict: The extractors can choose what strict mode means for them and related behavior. For strict mode, in certain situations, exception is raised by an exractor.

        Note:
            In strict mode, Exception is raised if the target object is empty (returns bool(obj) is False)
    '''

    def __init__(self, *, strict=True, target_types):
        self.__target_types = target_types
        self.__strict = strict

    @property
    def is_strict(self):
        '''
            Bool property that indicates whether this extractor is in strict mode or not.
        '''
        return self.__strict

    def extract(self, obj):
        '''
            Extracts objects from the given object as per concrete Extractor implementations.

            Arguments:
                obj: Extraction target.
        '''
        self.__validate(obj)  
        return self._extract(obj)           

    @abc.abstractmethod
    def _extract(self, obj):
        pass

    def __validate(self, obj):
        if type(obj) not in self.__target_types:
            raise Exception("Target object type mismatch. Provided {} of type {}. Should be one of {} for Extractor type: {}".format(obj, type(obj), self.__target_types, self.__class__.__name__))
        
        if self.is_strict and not obj:
            raise Exception("Provided target object is empty.")   


class _PositionExtractor(Extractor):
    '''
        Represents position based extractor objects for Tuples and Lists. Positions are considered in human counting (index + 1).

        Keyword Arguments:
            strict: If True, exception is raised if no element is found at provided position.
    '''

    def __init__(self, *, strict: bool):
        super().__init__(strict=strict, target_types={list, tuple})

    def __validate_seq_type(cls, sequence):
        if type(sequence) not in {tuple, list}:
            raise Exception(f"position extractors support only tuples and lists. Provided {sequence} of type {t}")

class _AtPositionExtractor(_PositionExtractor):
    '''
    Extracts object at a given position.

    Args:
        pos: (int) positions in an sequence

    Keyword Arguments:
        strict: If True, exception is raised if no element is found at provided position.
    ''' 
    def __init__(self, pos: int, strict=True):
        super().__init__(strict=strict)
        self.__index = pos - 1

    def _extract(self, sequence):
        return sequence[self.__index]

class _FirstPositionExtractor(_AtPositionExtractor):
    '''
    Extracts object at first position.

    Keyword Arguments:
        strict: If True, exception is raised if sequence is empty.
    ''' 

    def __init__(self, *, strict: bool=True):
        super().__init__(1, strict=strict)

class _LastPositionExtractor(_PositionExtractor):
    '''
    Extracts object at last position.

    Keyword Arguments:
        strict: If True, exception is raised if sequence is empty.
    ''' 

    def __init__(self, strict=True):
        super().__init__(strict=strict)
        self.__index = -1

    def _extract(self, sequence):
        return sequence[self.__index]

class _RandomPositionExtractor(_PositionExtractor):
    '''
    Extracts object at last position.

    Keyword Arguments:
        strict: If True, exception is raised if sequence is empty.
    ''' 

    def __init__(self, strict=True):
        super().__init__(strict=strict)
        self.__index = -1

    def _extract(self, sequence):
        return sequence[random.randint(0, len(sequence)-1)]

class _PositionsExtractor(Extractor):
    '''
        Represents positions based extractor objects for Tuples and Lists. Positions are considered in human counting (index + 1).

        Keyword Arguments:
            strict: If True, exception is raised if the extracted sequence is empty.
    '''

    def __init__(self, *, strict: bool):
        super().__init__(strict=strict, target_types={list, tuple})

    def _validated_output(self, sequence, out):
        if self.is_strict and not out:
            raise Exception("The extracted sequence is empty.")
        return out

class _AtPositionsExtractor(_PositionsExtractor):
    '''
        Represents fixed positions extractor object.

        Args:
            *pos: (int) positions in an sequence

        Keyword Arguments:
            strict: If True, exception is raised if for any of the provided positions, there is no entry in sequence. Default is True.
    ''' 

    def __init__(self, *pos, strict=True):
        super().__init__(strict=strict)
        self.__indices = [v-1 for v in pos]

    def _extract(self, sequence):
        out = []
        for i in self.__indices:
            try:
                out.append(sequence[i])
            except IndexError:
                if self.is_strict:
                    raise Exception("No entry found at pos {} in {}".format(i+1, sequence))
                else:
                    out.append(None)
        return self._validated_output(sequence, out)

class _SlicedPositionsExtractor(_PositionsExtractor):
    '''
        Represents sliced positions extractor object.

        Keyword Args:
            stop: Optional. Stop position. Default is end of sequence.
            start: Optional. Start position. Default is 1. 
            step: Optional. Position delta between two successive values.
            strict: If True, exception is raised if output sequence is empty. Default is True.

        Note:
            If not argument is provided, then all elements are included.
    ''' 

    def __init__(self, *, stop: int=None, start: int=1, step: int=1, strict: bool=True):
        super().__init__(strict=strict)
        if start is not None:
            start = start - 1
        # Python slices don't contain right boundary. So, no need to subtract 1 from stop.
        self.__slice = slice(start, stop, step)

    def _extract(self, sequence):
        return self._validated_output(sequence, sequence[self.__slice])


class _OddPositionsExtractor(_SlicedPositionsExtractor):
    '''
        Represents odd positions extracttor object which provides entries at odd positions.

        Keyword Args:
            strict: If True, exception is raised if output sequence is empty. Default is True.
    ''' 

    def __init__(self, *, strict: bool=True):
        super().__init__(start=1, step=2, strict=strict)

class _EvenPositionsExtractor(_SlicedPositionsExtractor):
    '''
        Represents odd positions extractor object which provides entries at even positions.

        Keyword Args:
            strict: If True, exception is raised if output sequence is empty. Default is True.
    ''' 

    def __init__(self, *, strict: bool=True):
        super().__init__(start=2, step=2, strict=strict)

class _RandomPositionsExtractor(_PositionExtractor):
    '''
    Extracts object at last position.

    Keyword Arguments:
        count: Number of unique random positions to be used.
        strict: If True, exception is raised if output sequence is empty or count is greater than sequence size. Default is True.
    ''' 
    def __init__(self, *, count, strict=True):
        super().__init__(strict=strict)
        self.__count = count

    def _extract(self, sequence):
        if len(sequence) < self.__count:
            if self.is_strict:
                raise Exception("Sequence {} is shorter than number of positions: {}".format(sequence, count))
            else:
                return [i for i in sequence]
        elif len(sequence) == self.__count:
            return [i for i in sequence]
        else:
            return random.sample(sequence, self.__count)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import base64

from arjuna.tpi.tracker import track

@track("trace")
class Image:
    '''
        An image object.

        Keyword Arguments:
            fpath: (Mandatory) Absolute path of the Image file.
            b64: Base64 representation of the image.
    '''

    def __init__(self, *, fpath: str, b64: str=None):
        self.__fpath = fpath
        self.__b64 = b64

    @property
    def file_name(self) -> str:
        '''
            File Name of screenshot file.
        '''

        return os.path.basename(self.__fpath)

    @property
    def full_path(self) -> str:
        '''
           Absolute path of screenshot file.
        '''

        return self.__fpath

    @property
    def base64(self) ->str:
        '''
            Base64 string for the image.
        '''
        if self.__b64:
            return self.__b64

        with open(self.__fpath, "rb") as f:
            content = f.read()
            return base64.b64encode(content)




# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from arjuna.tpi.helper.image import Image
from arjuna.tpi.tracker import track

@track("debug")
class ScreenShooter(metaclass=abc.ABCMeta):

    def take_screenshot(self, *, prefix: str=None) -> Image:
        '''
            Take screenshot.

            Args:
                prefix: prefix to be appended to the screenshot name.

            Returns:
                `Image` object.
        '''
        img = self._take_screenshot(prefix=prefix)
        if not isinstance(img, Image):
            raise Exception("_take_screenshot for ScreenShooter protocol must return an Image object.")
        return img

    @abc.abstractclassmethod
    def _take_screenshot(self, *, prefix: str=None) -> Image:
        pass


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import pytest
import unittest
from arjuna.core.utils import obj_utils
from arjuna.engine.data.record import DummyDataRecord
from arjuna.engine.data.store import SharedObjects
from arjuna.engine.meta import *
from typing import Callable
from arjuna.tpi.arjuna_types import ListOrTuple
from arjuna.tpi.helper.arjtype import CIStringDict
from arjuna.core.utils.obj_utils import get_function_meta_data
from arjuna.engine.selection.ref import *
from arjuna.tpi.error import TestDecoratorError

def _tc(cls):
    setattr(cls, 'get_test_qual_name', get_test_qual_name)
    return cls

def _call_func(func, request_wrapper, data=None, *args, **kwargs):
    from arjuna import log_info
    qual_name = request_wrapper.info.get_qual_name_with_data()
    log_info("Begin test function: {}".format(qual_name))  
    try:
        if data:      
            func(request=request_wrapper, data=data, *args, **kwargs)
        else:
            func(request=request_wrapper, *args, **kwargs)
    except Exception as e:
        log_info("End test function (with failure/error): {}. Exception Message: {}".format(qual_name, str(e)))    
        raise e
    log_info("End test function: {}".format(qual_name))


def __process_func_for_xfail(func, test_meta_data, xfail_obj):
    from arjuna import log_info
    if type(xfail_obj) is bool:
        if xfail_obj:
            return pytest.mark.xfail(True, reason="Expected Failure")(func)
        else:
            return func
    elif isinstance(xfail_obj, _XFail):
        return pytest.mark.xfail(
            xfail_obj.condition,
            reason = xfail_obj.reason,
            raises = xfail_obj.raises,
            run = xfail_obj.run,
            strict = xfail_obj.strict
            )(func)
    else:
        raise TestDecoratorError(func_qual_name=test_meta_data["info"]["qual_name"], msg="xfail argument should be of type XFail or bool.")


def __process_func_for_skip(func, test_meta_data, skip_obj):
    if type(skip_obj) is bool:
        if skip_obj:
            return pytest.mark.skip()(func)
        else:
            return func
    elif isinstance(skip_obj, _Skip):
        return pytest.mark.skipif(
            skip_obj.condition,
            reason = skip_obj.reason,
            )(func)
    else:
        raise TestDecoratorError(func_qual_name=test_meta_data["info"]["qual_name"], msg="skip argument should be of type Skip or bool.")          


def _simple_dec(func, test_meta_data):
    __process_test_meta_data(func, test_meta_data)
    from arjuna import Arjuna
    Arjuna.register_test_meta_data(test_meta_data['info']['qual_name'], CIStringDict(test_meta_data))
    func.__name__ = "check_" + func.__name__

    @functools.wraps(func)
    def wrapper(request, *args, **kwargs):
        request_wrapper = My(test_meta_data)
        request_wrapper.set_req_obj(request)
        _call_func(func, request_wrapper, *args, **kwargs)
    return wrapper

def _repr_record(record):
    return str(record)

def __process_test_meta_data(func, test_meta_data):
    func_meta_data = get_function_meta_data(func)
    qual_name = func_meta_data['qual_name']
    for name, value in test_meta_data['info'].items():
        if is_builtin_prop(name):
            expected_type = get_value_type(name)
            actual_type = type(value)
            if value is not None:
                if actual_type is not expected_type:
                    raise TestDecoratorError(qual_name, f">>{name}<< attr in @test should be of type >>{expected_type.__name__}<< but was provided value >>{value}<< of type >>{actual_type.__name__}<<")

            if name == "priority":
                if value not in set(range(1,6,1)):
                    raise TestDecoratorError(qual_name, f">>priority<< attr in @test can only accept 1/2/3/4/5 as value but was provided value >>{value}<<.")

    for container_name in {'tags', 'bugs', 'envs'}:
        container = test_meta_data[container_name]
        expected_types = {'set', 'list', 'tuple'}
        actual_type = type(container).__name__
        if actual_type not in expected_types:
            raise TestDecoratorError(qual_name, f">>{container_name}<< attr in @test can only be a set, list or tuple, but was provided value >>{value}<< of type >>{actual_type}<<")

        for entry in container:
            if type(entry) is not str:
                raise TestDecoratorError(qual_name, f">>{container_name}<< attr in @test can only contain strings, but was provided value >>{container}<< containing non-string entry >>{entry}<<.")

        test_meta_data[container_name] = set([s.strip().lower() for s in container])

    test_meta_data['info'].update(func_meta_data)

from collections import namedtuple

_XFail = namedtuple('XFail', "condition reason raises run strict")

_Skip = namedtuple('Skip', "reason condition")

def xfail(condition, *, reason: str="Expected Failure", raises: Exception=None, run: bool=True, strict: bool=False):
    '''
        Builder for an Expected Failure condition and decisions.
        Directly wraps **pytest.mark.xfail** https://docs.pytest.org/en/latest/reference.html#pytest-mark-xfail

        Args:
            condition: True/False. You can give a condition code evaluating to bool. You can give such code as a string as well. The string can be just a statement stating the reason.

        Keyword Arguments:
            reason: Why this test is expected to fail. Mandatory if condition is a bool.
            raises: Exception subclass expected to be raised by the test function (reported as XFailed); other exceptions will fail the test (reported as Failed).
            run: Should this test be executed. 
            strict: Test Suite will not be marked as a fail if this is False and this test passes or fails. If True, if this test passes, the test suite is marked as fail.
    '''
    return _XFail(
        condition = condition,
        reason = reason,
        raises = raises,
        run = run,
        strict = strict,
    )

def skip(condition: "bool or str", *, reason: str=None):
    '''
        Builder for a Skip condition.
        Wraps **pytest.mark.skip** and **pytest.mark.skipif**

        Args:
            condition: True/False. You can give a condition code evaluating to bool. You can give such code as a string as well. Or the string can be just a statement stating the reason.

        Keyword Arguments:
            reason: Why this test is to be skipped.
    '''
    return _Skip(
        condition = condition,
        reason = reason,
    )

def test(
            f:Callable=None, *, 
            id: str=None, 
            resources: ListOrTuple=None, 
            drive_with: 'DataSource'=None, 
            exclude_if: 'Relation'=None,
            xfail: "boolOrXFail"= False,
            skip: "boolOrSkip"= False,
            priority: int=5,
            author: str=None,
            idea: str=None,
            component: str=None,
            app_version: str='0.0.0',
            level: str=None,
            reviewed: bool= False,
            unstable: bool= False,
            tags: set=set(),
            bugs: set=set(),
            envs: set=set(),
            **test_attrs):
    '''
        Decorator for marking a function as a test function.

        Args:
            func: A Function with signature **f(request)**. The name request is mandatory and enforced.

        Keyword Arguments:
            id: Alnum string representing an ID which you want to associate with the test.
            resources: Fixtures/Resources that you want to associate this test with. Wraps pytest.mark.usefixtures. Instead of using this, you can also pass the names as direct arguments in the function signature.
            drive_with: Used for data driven testing. Argument can be Arjuna Data Source. Wraps **pytest.mark.parametrize**. If you use this argument, the test function signature must include a **data** argument e.g. 

                .. code-block:: python
                
                    @test(drive_with=<DS>)
                    def check_sample(request, data):
                        pass

            exclude_if: Define exclusion condition. Argument can be an Arjuna Relation. Wraps **pytest.mark.dependency**.
            xfail: Mark this test as a expected fail by setting to True. You can also use helper `xfail()` to create an advanced xfail construct. Wraps **pytest.mark.xfail**.
            skip: Mark this test as a expected skipped by setting to True. You can also use helper `skip()` to create an advanced skip construct. Wraps **pytest.mark.skip** and Wraps **pytest.mark.skipif**.
            priority: An integer value 1-5 depicting priority of this test, 1 being highest, 5 being lowest.
            author: Author of this test
            idea: The idea describing this test
            component: Primary software component that this test targets.
            app_version: Version of SUT that this test targets
            level: Level of this test.
            reviewed: Has this test been reviewed?
            unstable: Is this test unstable?
            tags: Set of tags for this test
            bugs: Set of bugs associated with this test
            envs: Set of Environment names on which this test is supposed to run.
            **test_attrs: Arbitrary name-value pairs to provide further test attributes.

        Note:
            The test function name must start with the prefix **check_**

            The test function must have the minimum signature as **check_<some_name>(request)** with **request** as the first argument.
    '''

    info_dict = CIStringDict()
    info_dict.update({
        'id': id,
        'priority': priority,
        'author': author,
        'idea': idea,
        'component': component,
        'app_version': app_version,
        'level': level,
        'reviewed': reviewed,
        'unstable': unstable,
    })

    info_dict.update({k.lower():v for k,v in test_attrs.items()})

    test_meta_data = {
        'info': info_dict,
        'tags': tags,
        'bugs': bugs,
        'envs': envs,
    }

    # Check if @test is provided without arguments
    if f is not None:
        return _simple_dec(f, test_meta_data)

    if resources:
        if type(resources) is str:
            resources = (resources)
        elif type(resources) is list:
            resources = tuple(resources)
        else:
            raise Exception("resources value must be a string or list/tuple of strings")

    def format_test_func(func):
        __process_test_meta_data(func, test_meta_data)
        func = __process_func_for_xfail(func, test_meta_data, xfail)
        func = __process_func_for_skip(func, test_meta_data, skip)
        from arjuna import Arjuna
        Arjuna.register_test_meta_data(test_meta_data['info']['qual_name'], CIStringDict(test_meta_data))
        orig_func = func
        if exclude_if:
            func = pytest.mark.dependency(name=id, depends=exclude_if())(func)
        else:
            func = pytest.mark.dependency(name=id)(func)

        if resources:
            func = pytest.mark.usefixtures(*resources)(func)

        if drive_with:
            records = drive_with.build().all_records
            func = pytest.mark.parametrize('data', records, ids=_repr_record)(func) 

        my = My(test_meta_data)

        @functools.wraps(orig_func)
        def wrapper_without_data(request, *args, **kwargs):
            my.set_req_obj(request)
            _call_func(func, my, *args, **kwargs)

        @functools.wraps(orig_func)
        def wrapper_with_data(request, data, *args, **kwargs):
            my.set_req_obj(request)
            _call_func(func, my, data, *args, **kwargs)

        if drive_with:
            return wrapper_with_data
        else:
            return wrapper_without_data
    
    return format_test_func

    




# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class NestedGuiElement:
    '''
        It is not an element container. It means it can not be find elements within itself.
    '''

    def __init__(self, automator, parent_element, wmd):
        pass

    



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.helper.arjtype import CIStringDict
from arjuna.interact.http.model.internal.processor.response import HttpExpectedResProcessor, HttpUnexpectedResProcessor

class HttpResponseYamlRepr:

    def __init__(self, session, resp_yaml):
        self.__xproc = None
        self.__unexproc = None
        
        if "unexpected" in resp_yaml:
            self.__unexproc = HttpUnexpectedResProcessor(session, CIStringDict(resp_yaml["unexpected"]))
            del resp_yaml["unexpected"]

        self.__xproc = HttpExpectedResProcessor(session, CIStringDict(resp_yaml))

    def validate(self, response):
        if self.__xproc:
            self.__xproc.validate(response)
        if self.__unexproc:
            self.__unexproc.validate(response)



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import os
import re
import argparse
import tempfile
import shutil
import logging
import traceback
import json
import sys
import abc

from arjuna.tpi.constant import *
from arjuna.core.constant import *
from arjuna.core.utils import sys_utils
from arjuna.core.utils import file_utils
from arjuna.core.reader.hocon import HoconFileReader, HoconConfigDictReader
from arjuna.configure.cli import CliArgsConfig
from .validation import *


class Command(metaclass=abc.ABCMeta):
    PENTRY = '''
    project.name = {}
    workspace.dir = "{}"
    '''

    def __init__(self):
        self.parser = None

    def print_help(self):
        self.parser.print_help()

    def _set_parser(self, parser):
        self.parser = parser

    def get_parser(self):
        return self.parser

    @abc.abstractmethod
    def execute(self, integrator, arg_dict):
        pass


class MainCommand(Command):

    def __init__(self):
        super().__init__()
        parser = argparse.ArgumentParser(prog='python -m arjuna', conflict_handler='resolve',
                                description="This is the CLI of Arjuna. Use the appropriate command and sub-commands as needed.")
        self._set_parser(parser)

    def create_subparsers(self):
        return self.parser.add_subparsers(title="Valid Commands", description="What do you want Arjuna to do?", dest='command')

    def convert_to_dict(self, args):
        def format_value(k, v):
            # if k in {"ao", "uo", "irules", "erules", 'ip', 'ep', }:
            #     return v
            if k == "report.formats" and v is None:
                v = ['XML', 'HTML']
            # if type(v) is list:
            #     return ",".join([str(i) for i in v])
            # else:
            return v
        try:
            args = self.parser.parse_args(args[1:])
            return {k:format_value(k,v) for k,v in vars(args).items()}
        except Exception as e:
            print("!!!Fatal Error!!!")
            print(e)
            import traceback
            traceback.print_exc()
            sys_utils.fexit()
        import sys
        sys.exit(1)

    def execute(self, arg_dict):
        pass


class FileObjectType(Enum):
    DIR = auto()
    FILE = auto()


class CreateProject(Command):

    COMMON_DIRS_FILES = (
        (FileObjectType.FILE, "__init__.py"),
        (FileObjectType.FILE, "pytest.ini"),
        (FileObjectType.DIR, "config"),
        (FileObjectType.FILE, "config/data.yaml"),
        (FileObjectType.FILE, "config/envs.yaml"),
        (FileObjectType.FILE, "config/groups.yaml"),
        (FileObjectType.FILE, "config/project.yaml"),
        (FileObjectType.FILE, "config/withx.yaml"),
        (FileObjectType.DIR, "data"),
        (FileObjectType.DIR, "data/source"),
        (FileObjectType.DIR, "data/reference"),
        (FileObjectType.DIR, "data/reference/contextual"),
        (FileObjectType.DIR, "data/reference/indexed"),
        (FileObjectType.DIR, "data/file"),
        (FileObjectType.DIR, "dbauto"),
        (FileObjectType.DIR, "dbauto/sql"),
        (FileObjectType.DIR, "dependency"),
        (FileObjectType.DIR, "l10n"),
        (FileObjectType.DIR, "guiauto"),
        (FileObjectType.DIR, "guiauto/driver"),
        (FileObjectType.DIR, "guiauto/driver/linux"),
        (FileObjectType.DIR, "guiauto/driver/mac"),
        (FileObjectType.DIR, "guiauto/driver/windows"),
        (FileObjectType.DIR, "guiauto/namespace"),
        (FileObjectType.DIR, "hook"),
        (FileObjectType.FILE, "hook/arjuna_config.py"),
        (FileObjectType.DIR, "report"),
        (FileObjectType.DIR, "script"),
        (FileObjectType.DIR, "test"),
        (FileObjectType.FILE, "test/__init__.py"),
        (FileObjectType.DIR, "test/pkg"),
        (FileObjectType.FILE, "test/pkg/__init__.py"),
        (FileObjectType.DIR, "tools"),
        (FileObjectType.DIR, "tools/bmproxy"),
        (FileObjectType.DIR, "lib"),
        (FileObjectType.FILE, "lib/__init__.py"),
        (FileObjectType.FILE, "lib/resource.py")
    )

    def __init__(self, subparsers, parents):
        super().__init__()
        self.parents = parents
        parser = subparsers.add_parser('create-project', parents=[parent.get_parser() for parent in parents], help="Create a new project")
        self._set_parser(parser)

    def __create_file_or_dir(self, project_temp_dir, ftype, frpath):
        if ftype == FileObjectType.DIR:
            os.makedirs(os.path.join(project_temp_dir, frpath))
        else:
            f = open(os.path.join(project_temp_dir, frpath), "w")
            f.close()

    def execute(self, arg_dict):
        def get_src_file_path(src):
            return os.path.join(os.path.dirname(os.path.realpath(__file__)), src)

        def get_proj_target_path(dest):
            return os.path.join(project_temp_dir, dest)

        def copy_file(src, dest):
            shutil.copyfile(get_src_file_path(src), get_proj_target_path(dest))

        for parent in self.parents:
            parent.process(arg_dict)
        # from arjuna import ArjunaCore
        pdir = arg_dict['project.root.dir']
        if os.path.exists(os.path.join(pdir, "config/project.yaml")):
            print("Arjuna project already exists at the specified location.")
            sys.exit(1)
        parent_dir = os.path.abspath(os.path.join(pdir, ".."))
        project_name = os.path.basename(pdir)
        with tempfile.TemporaryDirectory() as tdir:
            project_temp_dir = os.path.join(tdir, project_name)
            os.makedirs(project_temp_dir)
            for ftype, frpath in CreateProject.COMMON_DIRS_FILES:
                self.__create_file_or_dir(project_temp_dir, ftype, frpath)
            copy_file("../../res/project.yaml", "config/project.yaml")
            copy_file("../../res/check_dummy.txt", "test/pkg/check_dummy.py")
            copy_file("../../res/pt.ini", "pytest.ini")
            # copy_file("../../res/arjuna_launcher.py", "script/arjuna_launcher.py")
            # f = open(get_src_file_path("../../res/conftest.txt"), "r")
            # contents = f.read().format(project=project_name)
            # f.close()
            # f = open(get_proj_target_path("test/conftest.py"), "w")
            # f.write(contents)
            # f.close()
            for d in ["data/source", "data/file", "data/reference/contextual", "data/reference/indexed", "dbauto", "dbauto/sql", "dependency", "l10n", "guiauto/namespace", "tools/bmproxy"]:
                copy_file("../../res/placeholder.txt", d + "/placeholder.txt")
            for os_name in ["mac", "windows", "linux"]:
                copy_file("../../res/placeholder.txt", "guiauto/driver/{}/placeholder.txt".format(os_name))
            for f in os.listdir(project_temp_dir):
                try:
                    shutil.move(os.path.join(project_temp_dir, f), pdir)
                except Exception as e:
                    print(e)
                    pass

        print("Project {} successfully created at {}.".format(project_name, parent_dir))


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import os

from enum import Enum, auto
from abc import abstractmethod

from arjuna.tpi.constant import *
from arjuna.core.constant import *
from arjuna.tpi.error import GuiLabelNotPresentError
from arjuna.interact.gui.auto.finder.wmd import GuiWidgetMetaData
from arjuna.interact.gui.auto.finder._with import ImplWith

class FileFormat(Enum):
    # GNS = auto()
    # MGNS = auto()
    XLS = auto()
    XLSX = auto()
    YAML = auto()

class GuiNamespaceLoaderFactory:

    # Returns GuiNamespaceLoader
    @classmethod
    def create_namespace_loader(cls, config, ns_file_path):
        from arjuna.tpi.constant import ArjunaOption
        multi_context_enabled = config.value(ArjunaOption.GUIAUTO_DEF_MULTICONTEXT)
        context = multi_context_enabled and None or config.guiauto_context
        _, file_extension = os.path.splitext(ns_file_path)
        ext = file_extension.upper()[1:]
        considered_path = ns_file_path
        try:
            file_format = FileFormat[ext]
        except:
            raise Exception("Unsupported format for namespace: {}".format(file_extension))
        else:
            full_file_path = ns_file_path
            if os.path.isdir(full_file_path):
                raise Exception("Namespace file path is a directory and not a file: {}".format(considered_path))
            elif not os.path.isfile(full_file_path):
                from arjuna import log_warning
                log_warning("Namespace file path does not exist: {}".format(considered_path))
                return DummyGnsLoader(considered_path)
                # raise Exception()
            # if file_format == FileFormat.GNS:
            #     if multi_context_enabled:
            #         return MGNSFileLoader(full_file_path)
            #     else:
            #         return GNSFileLoader(full_file_path, context)
            if file_format == FileFormat.YAML:
                return YamlGnsLoader(full_file_path, context)
            else:
                raise Exception("Unsupported format for namespace: {}".format(file_extension))


class GuiNamespace:

    def __init__(self, name):
        self.__name = name
        # dict <string, dict<GuiAutomationContext, GuiWidgetMetaData>>
        self.__ns = {}

    def is_empty(self):
        return not self.__ns

    def add_element_meta_data(self, name, context, raw_locators, meta):
        from arjuna import log_debug
        log_debug("Loading {} label. Meta data: {}".format(name, str(meta)))
        wmd = GuiWidgetMetaData.create_wmd(*raw_locators, meta=meta)
        name = name.lower()
        if not self.has(name):
            self.__ns[name] = {}
        self.__ns[name][context] = wmd
        log_debug("Loaded {} label. EMD: {}".format(name, str(wmd)))

    def add_reference(self, name, value):
        self.__ns[name] = value

    def has(self, name):
        return name.lower() in self.__ns

    def has_context(self, name, context):
        if self.has(name):
            return context in self.__ns[name.lower()]
        return False

    # Needs to be thread-safe
    # Returns wmd for a context for a given gui name
    def get_meta_data(self, label, context):
        msg = ""
        if self.is_empty():
            msg = "Namespace is empty"
        if not self.has(label):
            raise GuiLabelNotPresentError(self.__name, label, msg=msg)
        elif not self.has_context(label, context):
            raise GuiLabelNotPresentError(self.__name, label, context, msg=msg)
        
        return self.__ns[label.lower()][context]

    @property
    def root_element_name(self):
        return self.__ns["__root__"]

    @property
    def anchor_element_name(self):
        return self.__ns["__anchor__"]

class BaseGuiNamespaceLoader:

    def __init__(self, name):
        self.__name = name
        self.__namespace = GuiNamespace(name)

    @property
    def name(self):
        return self.__name

    @property
    def namespace(self):
        return self.__namespace

    # Needs to be thread safe
    def add_element_meta_data(self, name, context, locators, meta):
        self.__namespace.add_element_meta_data(name, context, locators, meta)

    def add_reference(self, name, value):
        self.__namespace.add_reference(name, value)

    def _raise_notafile_exception(self, file_path):
        raise Exception("{} is not a file.".format(file_path))

    def _raise_filenotfound_exception(self, file_path):
        raise Exception("{} is not a valid file path.".format(file_path))

    def _raise_relativepath_exception(self, file_path):
        raise Exception("Gui namespace loader does not accept relative file path. {} is not a full file path.".format(file_path))

    def load(self):
        pass

class DummyGnsLoader(BaseGuiNamespaceLoader):

    def __init__(self, ns_file_path):
        super().__init__(os.path.basename(ns_file_path))

class YamlGnsLoader(BaseGuiNamespaceLoader):

    def __init__(self, ns_file_path, context):
        super().__init__(os.path.basename(ns_file_path))
        self.__context = context
        self.__ns_file = None
        self.__ns_path = None

        self.__ns = {}

        if not os.path.isabs(ns_file_path):
            super()._raise_relativepath_exception(ns_file_path)
        elif not os.path.exists(ns_file_path):
            super()._raise_filenotfound_exception(ns_file_path)
        elif not os.path.isfile(ns_file_path):
            super()._raise_notafile_exception(ns_file_path)

        self.__ns_path = ns_file_path
        self.__contexts = [context]

        self.__withx = None

        # self.__process()

    def load(self):
        
        from arjuna import Arjuna, log_debug
        from arjuna.configure.validator import Validator
        from arjuna.interact.gui.auto.finder._with import WithType
        from arjuna.tpi.parser.yaml import Yaml
        creation_context="Gui Namespace file at {}".format(self.__ns_path)
        yaml = Yaml.from_file(self.__ns_path, allow_any=True)

        if yaml is None: return

        if not yaml.has_section("labels"):
            # print("No labels configured. Skipping...")
            return

        from arjuna.interact.gui.auto.finder.withx import WithX
        if yaml.has_section("withx"):
            self.__withx = WithX(yaml.get_section("withx").as_map())
        else: 
            self.__withx = WithX()

        common_withx = Arjuna.get_withx_ref()

        from arjuna.tpi.error import GuiWidgetDefinitionError
        for label, label_map in yaml.get_section("labels").as_map().items():
            log_debug("Loading label: " + label)
            Validator.name(label)
            self.__ns[label.lower()] = {"locators" : {self.__context: []}, "meta": dict()}
            for entry in label_map:
                if type(label_map) is dict:
                    loc, loc_obj = entry, label_map[entry]
                elif type(label_map) is list:
                    if type(entry) is not dict or len(entry) != 1:
                        raise GuiWidgetDefinitionError("The GNS entry for label {} is not correctly formatted. For list content type, each list item should be a single item dictionary. Found: {}".format(label, label_map))
                    loc, loc_obj = list(entry.keys())[0], list(entry.values())[0]
                else:
                    raise GuiWidgetDefinitionError("The GNS entry for label {} is not correctly formatted. The content should either be a YAML mapping or YAML list. Found: {}".format(label, label_map))
                log_debug("Loading locator: " + loc)
                loc = loc.lower()
                wtype, wvalue = None, None
                if not self.__withx.has_locator(loc) and not common_withx.has_locator(loc):
                    wtype, wvalue = loc.upper(), loc_obj
                    if wtype in dir(WithType):
                        log_debug("Loading Arjuna defined Locator: " + loc)
                        if wtype in {'ATTR', 'FATTR', 'BATTR', 'EATTR'}:
                            if len(wvalue) > 1:
                                raise Exception("attr/fattr/battr/eattr entries in GNS should have a single key value pair mapping. Found: {} for locator type: {} for label: {}".format(wvalue, loc, label))
                            final_value = dict()
                            for k,v in wvalue.items():
                                final_value['name'] = k
                                final_value['value'] = v
                            wvalue = final_value
                        iloc = ImplWith(wtype=wtype, wvalue=wvalue, has_content_locator=False)
                        self.__ns[label.lower()]["locators"][self.__context].append(iloc)
                    else:
                        log_debug("Loading meta data for key: " + loc)
                        self.__ns[label.lower()]["meta"][wtype.lower()] = wvalue
                else:
                    if self.__withx.has_locator(loc):
                        wx = self.__withx
                    elif common_withx.has_locator(loc):
                        wx = common_withx
                    else:
                        raise Exception("No WithX locator with name {} found. Check GNS file at {}.".format(name, self.__ns_path))
                    try:
                        wtype, wvalue = wx.format(loc, loc_obj)
                    except Exception as e:
                        raise Exception("Error in implementation of withx locator extension: {} for label {}. Implementation: {}. Error: {}.".format(loc, label, wvalue, str(e)))

                    iloc = ImplWith(wtype=wtype, wvalue=wvalue, has_content_locator=False)
                    self.__ns[label.lower()]["locators"][self.__context].append(iloc)

            if not self.__ns[label.lower()]["locators"][self.__context]:
                raise Exception("No locators defined for label: {}".format(label))

        if yaml.has_section("load"):
            self.__load_targets = yaml.get_section("load").as_map()

            if "root" in self.__load_targets:
                self.__ns["__root__"] = self.__load_targets["root"].lower()
            else:
                self.__ns["__root__"] = None

            if "anchor" in self.__load_targets:
                self.__ns["__anchor__"] = self.__load_targets["anchor"].lower()
            else:
                self.__ns["__anchor__"] = None

        else:
            self.__ns["__root__"] = None
            self.__ns["__anchor__"] = None

        for ename, wmd in self.__ns.items():
            if ename not in {'__root__', '__anchor__'}:
                context_data = wmd["locators"]
                for context, locators in context_data.items():
                    self.add_element_meta_data(ename, context, locators, wmd["meta"])
                    log_debug("Loading {} label for {} context with locators: {} and meta {}.".format(ename, context, [str(l) for l in locators], wmd["meta"]))
        
        self.add_reference("__root__", self.__ns["__root__"])
        self.add_reference("__anchor__", self.__ns["__anchor__"])
